"""Class definition for ChatBot."""

import contextvars
from typing import Generator, Union


from llamabot.bot.simplebot import SimpleBot
from llamabot.config import default_language_model
from llamabot.components.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
)
from llamabot.components.history import History

prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class ChatBot(SimpleBot, History):
    """ChatBot that is primed with a system prompt, accepts a human message,
    and sends back a single response.

    It is essentially SimpleBot composed with History.

    :param system_prompt: The system prompt to use.
    :param session_name: The name of the session to use.
        This is used in the chat history class.
    :param temperature: The temperature to use.
    :param model_name: The model name to use.
    :param stream_target: The stream target to use.
        Should be one of ("stdout", "panel", "api").
    :param response_budget: The response budget to use, in terms of number of characters.
    :param completion_kwargs: Additional keyword arguments to pass to the completion function.
    """

    def __init__(
        self,
        system_prompt: str,
        session_name: str,
        stream_target: str = "stdout",
        temperature=0.0,
        model_name=default_language_model(),
        response_budget=2_000,
        **completion_kwargs,
    ):
        if stream_target == "api":
            raise ValueError("ChatBot does not support API streaming.")
        SimpleBot.__init__(
            self,
            system_prompt=system_prompt,
            temperature=temperature,
            model_name=model_name,
            stream_target=stream_target,
            **completion_kwargs,
        )
        History.__init__(self, session_name=session_name)
        self.response_budget = response_budget

    def __call__(self, message: str) -> Union[AIMessage, Generator]:
        """Call the ChatBot.

        :param human_message: The human message to use.
        :return: The response to the human message, primed by the system prompt.
        """
        human_message = HumanMessage(content=message)
        history = self.retrieve(
            query=human_message, character_budget=self.response_budget
        )
        messages = [self.system_prompt] + history + [human_message]
        match self.stream_target:
            case "stdout":
                response = self.stream_stdout(messages)
                self.append(human_message)
                self.append(response)
                return response
            case "panel":
                return self.stream_panel(messages)

        return AIMessage(content="")

    def __repr__(self):
        """Return a string representation of the ChatBot.

        :return: A string representation of the ChatBot.
        """
        representation = ""
        prefix = ""

        for message in self.messages:
            if isinstance(message, SystemMessage):
                prefix = "[System]\n"
            elif isinstance(message, HumanMessage):
                prefix = "[Human]\n"
            elif isinstance(message, AIMessage):
                prefix = "[AI]\n"

            representation += f"{prefix}{message.content}" + "\n\n"
        return representation



"""Git subcommand for LlamaBot CLI."""

import os
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Optional

import git
import pyperclip
import typer
from pydantic import BaseModel, Field, model_validator
from pyprojroot import here
from rich.console import Console

from llamabot import SimpleBot, prompt
from llamabot.bot.structuredbot import StructuredBot
from llamabot.code_manipulation import get_git_diff
from llamabot.config import default_language_model
from llamabot.prompt_library.git import (
    compose_git_activity_report,
    compose_release_notes,
)

gitapp = typer.Typer()


class CommitType(str, Enum):
    """Type of commit."""

    fix = "fix"
    feat = "feat"
    build = "build"
    chore = "chore"
    ci = "ci"
    docs = "docs"
    style = "style"
    refactor = "refactor"
    perf = "perf"
    test = "test"
    other = "other"


class DescriptionEntry(BaseModel):
    """Description entry."""

    txt: str = Field(
        ...,
        description="A single bullet point describing one major change in the commit.",
    )

    @model_validator(mode="after")
    def validate_description(self):
        """Validate description length."""
        if len(self.txt) > 160:
            raise ValueError(
                "Description should be less than or equal to 160 characters."
            )
        return self


class CommitMessage(BaseModel):
    """Commit message."""

    commit_type: CommitType = Field(
        ...,
        description=(
            "Type of change. Should usually be fix or feat. "
            "But others, based on the Angular convention, are allowed, "
            "such as build, chore, ci, docs, style, refactor, perf, test, and others."
        ),
    )
    scope: str = Field(
        ...,
        description=(
            "Scope of change. "
            "The scope in a conventional commit message represents "
            "a contextual identifier for the changes made in that particular commit. "
            "It provides a way to categorize and organize commits, "
            "making it easier to track and manage changes in a project. "
            "It can be thought of as a descriptor for the "
            "location, component, or module "
            "that is being modified or affected by the commit. "
            "For example, if a developer is working on a web application "
            "with multiple components such as the login system, "
            "user profile, and settings page, they can use scopes like "
            "login, profile or settings to indicate "
            "which part of the application the commit is related to. "
            "This helps other team members and future contributors understand "
            "the purpose of the commit "
            "without having to inspect the actual changes right away."
        ),
    )
    description: str = Field(
        ...,
        description=(
            "The description provides a concise summary of what the commit accomplishes, "
            "making it easier for team members and contributors to understand "
            "the purpose of the commit at a glance. "
            "It should be written in the present tense and be clear and informative. "
            "It should convey what the code changes do, "
            "rather than how they were implemented. "
            "It is not necessary to provide every detail in the description; "
            "instead, focus on the high-level overview of the modifications."
        ),
    )

    body: list[DescriptionEntry] = Field(
        ...,
        description=(
            "Unlike the description, "
            "which provides a concise summary of the changes made in the commit, "
            "the body allows for a more detailed explanation of the commit. "
            "The body section is used to provide additional context, reasoning, "
            "or implementation details that might be helpful "
            "for other developers, reviewers, or future contributors "
            "to understand the commit thoroughly. "
            "It is particularly useful "
            "when the changes introduced in the commit are complex "
            "or when there are specific design decisions that need to be explained. "
            "It is worth mentioning that having a body section is optional, "
            "and not every commit requires it. "
            "Simple and straightforward changes "
            "might not need a detailed explanation beyond the description."
            "If provided, this should be bullet points."
        ),
    )

    breaking_change: bool = Field(
        ..., description="Whether or not there is a breaking change in the commit. "
    )

    footer: str = Field("", description="An optional footer.")
    emoji: str = Field(..., description="An emoji that represents the commit content.")

    @model_validator(mode="after")
    def validate_body(self):
        """Validate the body length."""
        if len(self.body) > 10:
            raise ValueError("Description entries should be no more than 10 in length.")
        return self

    def format(self) -> str:
        """Format the commit message according to the provided model.

        :return: Formatted commit message as a string.
        """
        return _fmt(self)


@prompt(role="system")
def _fmt(cm) -> str:
    """{{ cm.commit_type.value }}({{ cm.scope }}){{ cm.emoji }}{%if cm.breaking_change %}!{% else %}{% endif %}: {{ cm.description }}

    {% for bullet in cm.body %}- {{ bullet.txt }}
    {% endfor %}

    {% if cm.footer %}{{ cm.footer }}{% endif %}
    """  # noqa: E501


def commitbot(model_name: str = default_language_model()) -> StructuredBot:
    """Return a structured bot for writing commit messages."""

    @prompt(role="system")
    def commitbot_sysprompt() -> str:
        """You are an expert software developer
        who writes excellent and accurate commit messages.
        You are going to be given a diff as input,
        and you will generate a structured JSON output
        based on the pydantic model provided.
        Ensure that your commit message is formatted as a conventional commit message.
        """

    bot = StructuredBot(
        system_prompt=commitbot_sysprompt(),
        pydantic_model=CommitMessage,
        model_name=model_name,
        stream_target="none",
    )
    return bot


@gitapp.command()
def hooks(model_name: str = default_language_model()):
    """Install a commit message hook that runs the commit message through the bot.

    :raises RuntimeError: If the current directory is not a git repository root.
    """
    # Check that we are in a repository's root. There should be a ".git" folder.
    # Use pathlib to verify.
    if not Path(".git").exists():
        raise RuntimeError(
            "You must be in a git repository root folder to use this command. "
            "Please `cd` into your git repo's root folder and try again, "
            "or use `git init` to create a new repository (if you haven't already)."
        )

    with open(".git/hooks/prepare-commit-msg", "w+") as f:
        contents = f"""#!/bin/sh

echo "Script started with arguments: $@"

# Check if the arguments contain '-m'
commit_message_provided=false

for arg in "$@"; do
    if [ "$arg" = "message" ]; then
        commit_message_provided=true
        break
    fi
done

if $commit_message_provided; then
    echo "Commit message provided, skipping llamabot git compose."
else
    echo "No commit message found, running llamabot git compose."
    llamabot git compose --model-name {model_name}
fi
"""
        f.write(contents)
    os.chmod(".git/hooks/prepare-commit-msg", 0o755)
    typer.echo("Commit message hook successfully installed! 🎉")


@gitapp.command()
def compose(model_name: str = default_language_model()):
    """Autowrite commit message based on the diff."""
    try:
        diff = get_git_diff()
        bot = commitbot(model_name)
        response = bot(diff, verbose=True)
        print(response.format())
        with open(".git/COMMIT_EDITMSG", "w") as f:
            f.write(response.format().content)
    except Exception as e:
        typer.echo(f"Error encountered: {e}", err=True)
        typer.echo("Please write your own commit message.", err=True)
        raise e


@gitapp.command()
def write_release_notes(release_notes_dir: Path = Path("./docs/releases")):
    """Write release notes for the latest two tags to the release notes directory.

    :param release_notes_dir: The directory to write the release notes to.
        Defaults to "./docs/releases".
    """
    repo = git.Repo(here())
    tags = sorted(repo.tags, key=lambda t: t.commit.committed_datetime)
    if len(tags) == 0:
        # No tags, get all commit messages from the very first commit
        log_info = repo.git.log()
    elif len(tags) == 1:
        # Only one tag, get all commit messages from that tag to the current commit
        tag = tags[0]
        log_info = repo.git.log(f"{tag.commit.hexsha}..HEAD")
    else:
        # More than one tag, get all commit messages between the last two tags
        tag1, tag2 = tags[-2], tags[-1]
        log_info = repo.git.log(f"{tag1.commit.hexsha}..{tag2.commit.hexsha}")

    console = Console()
    bot = SimpleBot(
        "You are an expert software developer "
        "who knows how to write excellent release notes based on git commit logs.",
        model_name=default_language_model(),
        stream_target="none",
    )
    with console.status("[bold green]Generating release notes...", spinner="dots"):
        notes = bot(compose_release_notes(log_info))

    # Create release_notes_dir if it doesn't exist:
    release_notes_dir.mkdir(parents=True, exist_ok=True)
    # Ensure only one newline at the end of the file
    trimmed_notes = notes.content.rstrip() + "\n"

    # Write release notes to the file
    with open(release_notes_dir / f"{tag2.name}.md", "w+") as f:
        f.write(trimmed_notes)


@gitapp.command()
def report(
    hours: Optional[int] = typer.Option(None, help="The number of hours to report on."),
    start_date: Optional[str] = typer.Option(
        None, help="The start date to report on. Format: YYYY-MM-DD"
    ),
    end_date: Optional[str] = typer.Option(
        None, help="The end date to report on. Format: YYYY-MM-DD"
    ),
    model_name: str = default_language_model(),
):
    """
    Write a report on the work done based on git commit logs.

    If hours is provided, it reports on the last specified hours.
    If start_date and end_date are provided, it reports on that date range.
    If neither is provided, it raises an error.

    :param hours: The number of hours to report on.
    :param start_date: The start date to report on.
    :param end_date: The end date to report on.
    :param model_name: The model name to use.
        Consult LiteLLM's documentation for options.
    """
    repo = git.Repo(here())

    if hours is not None:
        now = datetime.now()
        time_ago = now - timedelta(hours=hours)
        now_str = now.strftime("%Y-%m-%dT%H:%M:%S")
        time_ago_str = time_ago.strftime("%Y-%m-%dT%H:%M:%S")
    elif start_date and end_date:
        time_ago_str = start_date
        now_str = end_date
    else:
        raise ValueError(
            "Either 'hours' or both 'start_date' and 'end_date' must be provided."
        )

    log_info = repo.git.log(f"--since={time_ago_str}", f"--until={now_str}")
    bot = SimpleBot(
        "You are an expert software developer who writes excellent reports based on git commit logs.",
        model_name=model_name,
        stream_target="none",
    )

    console = Console()
    with console.status("[bold green]Generating report...", spinner="dots"):
        report = bot(
            compose_git_activity_report(
                log_info, str(hours) or f"from {start_date} to {end_date}"
            )
        )

    print(report.content)
    # Copy report content to clipboard
    pyperclip.copy(report.content)
    typer.echo("Report copied to clipboard. Paste it wherever you need!")



"""Tutorial bot and prompts."""

from llamabot import SimpleBot
from llamabot.prompt_manager import prompt


@prompt(role="system")
def tutorialbot_system_prompt():
    """
    You are an expert python tutorial writer.

    Your style is engaging without being too entertaining.

    You will be provided a library of code.
    Your task is to write a tutorial for the code.

    Do not copy code from the library; instead,
    simply presume it is imported.

    Where possible, show how the code is used in relation to one another.
    For example, the output of one function being used in a downstream function.
    Do not hallucinate code that does not exist.
    """


def tutorialbot() -> SimpleBot:
    """Return a tutorial bot.

    :return: Tutorial bot
    """
    return SimpleBot(tutorialbot_system_prompt())


@prompt(role="user")
def module_tutorial_writer(source_file):
    """Please help me write a tutorial for the following code.

    {{ source_file }}

    It should be in Markdown format.

    # noqa: DAR101
    """


@prompt(role="user")
def typer_cli_tutorial(source_file, additional_notes):
    """I have the following CLI source file.

    {{ source_file }}

    It is a Typer CLI module, therefore, the commands are kebab-cased
    (like this: `<clitool> some-command`),
    where `<clitool>` is replaced by the actual command line tool name.
    Please help me write a tutorial about it.
    Please provide exhaustive examples about the combinations of command arguments
    that can be used.

    I have additional notes that you can use:

    {{ additional_notes }}

    [TUTORIAL BEGIN]  # noqa: DAR101
    """



"""Launch the web app to visualize and compare prompts and messages."""

import typer
import uvicorn
from pathlib import Path
from pyprojroot import here
from llamabot.web.app import create_app

app = typer.Typer()


@app.command()
def launch(db_path: Path = here() / "message_log.db"):
    """Launch the web app to visualize and compare prompts and messages.

    :param db_path: The path to the database to use.
    """

    fastapi_app = create_app(db_path)
    uvicorn.run(fastapi_app, host="0.0.0.0", port=8000)


if __name__ == "__main__":
    app()



"""Generic document store for LlamaBot.

The interface that we need is implemented here:

- append,
- extend,
- retrieve

ChromaDB is a great default choice because of its simplicity and FOSS nature.
Hence we use it by default.
"""

from hashlib import sha256
from pathlib import Path
from typing import Callable


from tqdm.auto import tqdm

from llamabot.doc_processor import magic_load_doc, split_document


class AbstractDocumentStore:
    """Abstract document store for LlamaBot."""

    def __init__(self):
        raise NotImplementedError()

    def append(self, document: str):
        """Append a document to the store.

        :param document: The document to append.
        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def extend(self, documents: list[str]):
        """Extend a list of documents to the store.

        :param documents: The documents to append.
        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve a list of documents from the store.

        :param query: The query to make of the documents.
        :param n_results: The number of results to retrieve.
        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def reset(self):
        """Reset the document store.

        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def add_documents(
        self,
        document_paths: Path | list[Path],
        chunk_size: int = 2_000,
        chunk_overlap: int = 500,
    ):
        """Add documents to the QueryBot DocumentStore.

        :param document_paths: The document paths to add to the store.
        :param chunk_size: The size of each chunk.
        :param chunk_overlap: The amount of overlap between chunks.
        """
        if isinstance(document_paths, Path):
            document_paths = [document_paths]

        for document_path in tqdm(document_paths):
            document = magic_load_doc(document_path)
            splitted_document = split_document(
                document, chunk_size=chunk_size, chunk_overlap=chunk_overlap
            )
            self.extend(splitted_document)
        self.__post_add_documents__()

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        raise NotImplementedError()


class ChromaDBDocStore(AbstractDocumentStore):
    """A document store for LlamaBot that wraps around ChromaDB."""

    def __init__(
        self,
        collection_name: str,
        storage_path: Path = Path.home() / ".llamabot" / "chroma.db",
    ):
        import chromadb

        client = chromadb.PersistentClient(path=str(storage_path))
        collection = client.create_collection(collection_name, get_or_create=True)

        self.storage_path = storage_path
        self.client = client
        self.collection = collection
        self.collection_name = collection_name
        self.existing_records = collection.get()

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        pass

    def append(self, document: str, metadata: dict = {}):
        """Append a document to the store.

        :param document: The document to append.
        """
        doc_id = sha256(document.encode()).hexdigest()

        add_kwargs = dict(
            documents=document,
            ids=doc_id,
        )
        if metadata:
            add_kwargs["metadatas"] = metadata

        self.collection.add(**add_kwargs)

    def extend(self, documents: list[str]):
        """Extend the document store.

        :param documents: Iterable of documents.
        """
        for document in documents:
            self.append(document)

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve documents from the store.

        :param query: The query to use to retrieve documents.
        """
        import chromadb

        # Use Vectordb to get documents.
        results: chromadb.QueryResult = self.collection.query(
            query_texts=query, n_results=n_results
        )
        vectordb_documents: list[str] = results["documents"][0]
        return vectordb_documents

        # Return the union of the retrieved documents
        # union = set(vectordb_documents).union(bm25_documents)
        # return list(union)

    def reset(self):
        """Reset the document store."""
        self.client.delete_collection(self.collection_name)
        self.collection = self.client.create_collection(
            self.collection_name, get_or_create=True
        )


class LanceDBDocStore(AbstractDocumentStore):
    """A document store for LlamaBot that wraps around LanceDB."""

    def __init__(
        self,
        table_name: str,
        storage_path: Path = Path.home() / ".llamabot" / "lancedb",
    ):
        import lancedb
        from lancedb.embeddings import EmbeddingFunctionRegistry, get_registry
        from lancedb.pydantic import LanceModel, Vector

        registry: EmbeddingFunctionRegistry = get_registry()
        func: Callable = registry.get(name="sentence-transformers").create()

        class DocstoreEntry(LanceModel):
            """LanceDB DocumentStore Entry."""

            document: str = func.SourceField()
            vector: Vector(func.ndims()) = func.VectorField()

        self.schema = DocstoreEntry
        self.table_name = table_name
        self.db = lancedb.connect(storage_path)

        try:
            self.table = self.db.open_table(table_name)
        except FileNotFoundError:
            self.table = self.db.create_table(table_name, schema=self.schema)

        try:
            self.existing_records = [
                item.document
                for item in self.table.search().limit(None).to_pydantic(self.schema)
            ]
        except ValueError:
            self.existing_records = []

    def __contains__(self, other: str) -> bool:
        """Returns boolean whether the 'other' document is in the store.

        :param other: The document to search for.
        :return: True if the document is in the store, False otherwise.
        """
        all_items = self.table.search().limit(None).to_pydantic(self.schema)
        texts = set([item.document for item in all_items])
        return other in texts

    def append(self, document: str):
        """Append a document to the store.

        :param document: The document to append.
        """
        # Avoid duplication of documents in LanceDB.
        if document not in self.existing_records:
            self.table.add([{"document": document}])
            self.existing_records.append(document)

    def extend(self, documents: list[str]):
        """Extend a list of documents to the store.

        :param documents: The documents to append.
        """
        # self.table.add(documents)
        for doc in tqdm(documents):
            self.append(doc)

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve a list of documents from the store.

        :param query: The query to use to retrieve documents.
        :param n_results: The number of results to retrieve.
        :return: A list of documents.
        """
        results: list = (
            self.table.search(query, query_type="hybrid")
            .limit(n_results)
            .to_pydantic(self.schema)
        )
        return [r.document for r in results]

    def reset(self):
        """Reset the document store."""
        self.db.drop_table(self.table_name)
        self.table = self.db.create_table(self.table_name, schema=self.schema)

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        self.table.create_fts_index("document", replace=True)


class BM25DocStore(AbstractDocumentStore):
    """In-memory BM25-based retrieval document store."""

    def __init__(self):
        self.documents: list = []

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        pass

    def append(self, document: str):
        """Append a document to the store.

        :param document: The document to append.
        """
        self.documents.append(document)

    def extend(self, documents: list[str]):
        """Extend a list of documents to the store.

        :param documents: The documents to append.
        """
        self.documents.extend(documents)

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve documents from the store.

        :param query: The query to use to retrieve documents.
        :param n_results: The number of results to retrieve.
        :return: A list of documents.
        """
        from rank_bm25 import BM25Okapi

        # Use BM25 to get documents.
        tokenized_docs = [doc.split() for doc in self.documents]
        search_engine = BM25Okapi(tokenized_docs)
        docs: list[str] = search_engine.get_top_n(
            query.split(), self.documents, n=n_results
        )
        return docs

    def reset(self):
        """Reset the document store."""
        self.documents = []



"""CLI utility functions."""

import getpass
from pathlib import Path

import typer
from prompt_toolkit import prompt
from prompt_toolkit.formatted_text import HTML


def get_valid_input(prompt, valid_inputs=("y", "n")):
    """
    This function prompts the user for input and validates it.

    .. code-block:: python

        user_choice = get_valid_input("Enter 'y' for yes or 'n' for no: ")

    :param prompt: The prompt to display to the user.
    :param valid_inputs: A tuple of valid inputs.
    :return: The validated user input, either 'y' or 'n'.
    """
    msg = f"{prompt} {valid_inputs}"
    while True:
        user_input = input(msg).lower()
        if user_input in valid_inputs:
            return user_input
        else:
            print(f"Invalid input. Please enter one of { '/'.join(valid_inputs) }.")


def configure_environment_variable(env_var: str, env_value: str):
    """Configure environment variable in ~/.llamabotrc.

    :param env_var: The name of the environment variable.
        Will be converted to upper case.
    :param env_value: The API key.
    """
    config_file = Path.home() / ".llamabot/.llamabotrc"
    env_var_line = f'export {env_var.upper()}="{env_value}"'

    if config_file.exists():
        with open(config_file, "r") as file:
            content = file.readlines()

        found_env_var = False
        with open(config_file, "w") as file:
            for line in content:
                if f"export {env_var.upper()}=" in line:
                    file.write(env_var_line + "\n")
                    found_env_var = True
                else:
                    file.write(line)

        if not found_env_var:
            with open(config_file, "a") as file:
                file.write(env_var_line + "\n")
    else:
        with open(config_file, "w") as file:
            file.write(env_var_line + "\n")


def uniform_prompt():
    """The uniform prompt for all llamabot commands.

    :return: The prompt, partialled out and ready to accept user input."""

    def bottom_toolbar():
        """The bottom toolbar for the prompt.

        :return: The bottom toolbar.
        """
        return HTML(
            " Multi-line input is enabled. "
            "Use Meta+Enter or Escape->Enter to submit. "
            "Type 'exit' or 'quit' to exit, or else use Ctrl+C. "
        )

    return prompt(
        f"[{getpass.getuser()}]: ", multiline=True, bottom_toolbar=bottom_toolbar
    )


def exit_if_asked(query: str):
    """Check if the user wants to exit.

    If yes, exit the program.

    :param query: The user's query.
    :raises Exit: If the user types "exit" or "quit".
    """
    query = query.strip(" ").strip("\n").lower()
    if query in ["exit", "quit"]:
        print("It was fun chatting! Have a great day!")
        raise typer.Exit(0)



"""Class for storing and retrieving messages from a message history.

Message history is always stored in a chroma database.
We can retrieve the last N messages from the database
or use vector similarity search to identify similar messages to retrieve (RAGHistory).
"""

from pathlib import Path
from hashlib import sha256
from llamabot.components.messages import (
    BaseMessage,
    RetrievedMessage,
    retrieve_messages_up_to_budget,
)


class History:
    """History of messages."""

    def __init__(self, session_name: str):
        self.messages: list[BaseMessage] = []
        self.session_name = session_name

    def append(self, message: BaseMessage):
        """Append a message to the history."""
        self.messages.append(message)

    def retrieve(self, query: BaseMessage, character_budget: int) -> list[BaseMessage]:
        """Retrieve messages from the history up to the the character budget.

        We use the character budget in order to simplify how we retrieve messages.

        :param query: The query to use to retrieve messages. Not used in this class.
        :param character_budget: The number of characters to retrieve.
        """
        return retrieve_messages_up_to_budget(self.messages, character_budget)

    def __getitem__(self, index):
        """Get the message at the given index."""
        return self.messages[index]


class RAGHistory:
    """History of messages with vector similarity enabled."""

    def __init__(
        self,
        session_name: str,
        db_path: Path = Path.home() / ".llamabot" / "chroma.db",
    ):
        import chromadb

        client = chromadb.PersistentClient(path=str(db_path))
        collection = client.create_collection(session_name, get_or_create=True)

        self.db_path = db_path
        self.client = client
        self.collection = collection
        self.session_name = session_name

    def append(self, message: BaseMessage):
        """Append a message to the history.

        :param message: The message to append.
        """
        doc_id = sha256(message.content.encode()).hexdigest()
        self.collection.add(documents=message.content, ids=doc_id)

    def retrieve(
        self, query: BaseMessage, character_budget: int, n_results: int = 50
    ) -> list[BaseMessage]:
        """Retrieve messages from the history up to the the character budget.

        :param query: The query to use to retrieve messages.
        :param character_budget: The number of characters to retrieve.
        :param n_results: The number of results to retrieve.
        :returns: The retrieved messages.
        """
        results: dict = self.collection.query(
            query_texts=query.content, n_results=n_results
        )
        messages: list[RetrievedMessage] = [
            RetrievedMessage(content=doc) for doc in results["documents"][0]
        ]
        return retrieve_messages_up_to_budget(messages, character_budget)

    def __getitem__(self, index) -> RetrievedMessage:
        """Get the message at the given index.

        :param index: The index to use.
        :returns: The retrieved message.
        """
        documents = self.collection.get()["documents"]
        document = RetrievedMessage(content=documents[index])
        return document



"""
This module provides a llamabot global cache for caching responses from bots.
"""

from diskcache import Cache
from pathlib import Path
import os

CACHE_DIR = Path.home() / ".llamabot" / "cache"
CACHE_DIR.mkdir(parents=True, exist_ok=True)

# Get cache timeout from environment variable, default to 1 day (86400 seconds)
CACHE_TIMEOUT = int(os.getenv("LLAMABOT_CACHE_TIMEOUT", 86400))

cache = Cache(CACHE_DIR, timeout=CACHE_TIMEOUT)



"""Definitions for the different types of messages that can be sent."""

from pydantic import BaseModel, Field


class BaseMessage(BaseModel):
    """A base message class."""

    role: str
    content: str
    prompt_hash: str | None = Field(default=None)

    # Implement slicing for message contents so that I can get content[:-i].
    def __getitem__(self, index):
        """Get the content of the message at the given index."""
        return self.__class__(content=self.content[index], role=self.role)

    def __len__(self):
        """Get the length of the message."""
        return len(self.content)


class SystemMessage(BaseMessage):
    """A message from the system."""

    content: str
    role: str = "system"


class HumanMessage(BaseMessage):
    """A message from a human."""

    content: str
    role: str = "user"


class AIMessage(BaseMessage):
    """A message from the AI."""

    content: str
    role: str = "assistant"


class ToolMessage(BaseMessage):
    """A message from the AI."""

    content: str
    role: str = "tool"


class RetrievedMessage(BaseMessage):
    """A message retrieved from the history."""

    content: str
    role: str = "system"


def retrieve_messages_up_to_budget(
    messages: list[BaseMessage], character_budget: int
) -> list[BaseMessage]:
    """Retrieve messages up to the character budget.

    :param messages: The messages to retrieve.
    :param character_budget: The character budget to use.
    :returns: The retrieved messages.
    """
    used_chars = 0
    retrieved_messages = []
    for message in messages:
        if not isinstance(message, (BaseMessage, str)):
            raise ValueError(
                f"Expected message to be of type BaseMessage or str, got {type(message)}"
            )
        used_chars += len(message)
        if used_chars > character_budget:
            # append whatever is left
            retrieved_messages.append(message[: used_chars - character_budget])
            break
        retrieved_messages.append(message)
    return retrieved_messages



"""Decorators for turning Python functions into Jinja2-templated prompts.

This module contains decorators for turning Python functions' docstrings
into Jinja2-templated functions that accept function arguments
and return the Jinja2 templated docstrings as strings when called.

Inspired from the Outlines library.
"""

from functools import wraps
import jinja2
from jinja2 import meta
import inspect
from textwrap import dedent
from llamabot.recorder import store_prompt_version
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from pyprojroot import here
from llamabot.recorder import Base, upgrade_database
from llamabot.components.messages import BaseMessage
from typing import Literal


def version_prompt(template: str, function_name: str) -> str:
    """Version a prompt template and return its hash.

    :param template: The prompt template to version.
    :param function_name: The name of the function being decorated.
    :return: The hash of the prompt template.
    """
    engine = create_engine(f"sqlite:///{here() / 'message_log.db'}")
    Base.metadata.create_all(engine)
    upgrade_database(engine)
    Session = sessionmaker(bind=engine)
    session = Session()

    try:
        stored_prompt = store_prompt_version(session, template, function_name)
        return stored_prompt.hash
    finally:
        session.close()


def prompt(role: Literal["system", "user", "assistant"] = "system"):
    """Wrap a Python function into a Jinja2-templated prompt with version control.

    :param role: The role of the prompt.
    :return: The prompt decorator.
    """

    def decorator(func):
        """Decorator function.

        :param func: The function to wrap.
        :return: The wrapped function.
        """

        @wraps(func)
        def wrapper(*args, **kwargs) -> BaseMessage:
            """Wrapper function.

            :param args: Positional arguments to the function.
            :param kwargs: Keyword arguments to the function.
            :return: A BaseMessage with the Jinja2-templated docstring as content.
            :raises ValueError: If a variable in the docstring
                is not passed into the function.
            """
            # get the function's docstring.
            docstring = func.__doc__

            # map args and kwargs onto func's signature.
            signature = inspect.signature(func)
            kwargs = signature.bind(*args, **kwargs).arguments

            # create a Jinja2 environment
            env = jinja2.Environment()

            # parse the docstring
            parsed_content = env.parse(docstring)

            # get all variables in the docstring
            variables = meta.find_undeclared_variables(parsed_content)

            # check if all variables are in kwargs
            for var in variables:
                if var not in kwargs:
                    raise ValueError(
                        f"Variable '{var}' was not passed into the function"
                    )

            # Version the prompt template
            prompt_hash = version_prompt(docstring, func.__name__)

            # interpolate docstring with args and kwargs
            template = jinja2.Template(docstring)
            string = template.render(**kwargs)

            # dedent the string
            # Split the string into lines
            lines = string.split("\n")

            # Dedent each line
            dedented_lines = [dedent(line) for line in lines]

            # Join the lines back into a single string
            dedented_string = "\n".join(dedented_lines).strip()

            # Return a BaseMessage with the specified role
            return BaseMessage(
                role=role, content=dedented_string, prompt_hash=prompt_hash
            )

        return wrapper

    return decorator



"""Implementation source code for PydanticBot.

Courtesey of Elliot Salisbury (@ElliotSalisbury).

Highly inspired by instructor by jxnl (https://github.com/jxnl/instructor).
"""

from typing import Union
from loguru import logger

from llamabot import SimpleBot
from llamabot.components.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
)
from pydantic import BaseModel, ValidationError
from llamabot.config import default_language_model
from llamabot.prompt_manager import prompt


@prompt(role="user")
def bot_task(schema: str) -> BaseMessage:
    """Your task is to return the data in a json object
    that matches the following json_schema:

    {{ schema }}

    Only return an INSTANCE of the schema, do not return the schema itself.
    """


class StructuredBot(SimpleBot):
    """StructuredBot is given a Pydantic Model and expects the LLM to return
    a JSON structure that conforms to the model schema.
    It will validate the returned json against the pydantic model,
    prompting the LLM to fix any of the validation errors if it does not validate,
    and then explicitly return an instance of that model.

    StructuredBot streams only to stdout; other modes of streaming are not supported.

    This is distinct from SimpleBot's JSON-mode behaviour in the following ways:

    1. JSON mode ensures a valid JSON is returned, but it does not guarantee it's in the schema you requested.
    2. We can customize the validation rules to our needs. For example: checking for hallucinations.
    3. Explicitly controllable number of retries.
    4. Directly returning an instance of the Pydantic model rather than just a plain JSON string.
    """

    def __init__(
        self,
        system_prompt: str,
        pydantic_model: BaseModel,
        model_name: str = default_language_model(),
        stream_target: str = "stdout",
        **completion_kwargs,
    ):
        super().__init__(
            system_prompt,
            stream_target=stream_target,
            json_mode=True,
            model_name=model_name,
            **completion_kwargs,
        )

        self.pydantic_model = pydantic_model

    def task_message(self) -> HumanMessage:
        """Compose instructions for what the bot is supposed to do."""
        schema = self.pydantic_model.model_json_schema()
        return bot_task(schema)

    def get_validation_error_message(self, exception: ValidationError) -> HumanMessage:
        """Return a validation error message to feed back to the bot.

        :param exception: The raised ValidationError from pydantic.
        """
        exception_msg = exception.json()
        return HumanMessage(
            content=f"There was a validation error: {exception_msg} try again."
        )

    def _extract_json_from_response(self, response: AIMessage):
        """Extract JSON content from the LLM response.

        :param response: The LLM response message.
        """
        content = response.content
        first_paren = content.find("{")
        last_paren = content.rfind("}")
        return content[first_paren : last_paren + 1]

    def __call__(
        self,
        message: Union[str, BaseMessage],
        num_attempts: int = 10,
        verbose: bool = False,
    ) -> BaseModel | None:
        """Process the input message and return an instance of the Pydantic model.

        :param message: The text on which to parse to generate the structured response.
        """
        if isinstance(message, str):
            message = HumanMessage(content=message)

        messages = [
            self.system_prompt,
            self.task_message(),
            message,
        ]

        # we'll attempt to get the response from the model and validate it
        for attempt in range(num_attempts):
            try:
                match self.stream_target:
                    case "stdout":
                        response = self.stream_stdout(messages)
                    case "none":
                        response = self.stream_none(messages)

                # parse the response, and validate it against the pydantic model
                codeblock = self._extract_json_from_response(response)
                obj = self.pydantic_model.model_validate_json(codeblock)
                return obj

            except ValidationError as e:
                # we're on our last try, so we raise the error
                if attempt == num_attempts:
                    raise e

                # Otherwise, if we failed, give the LLM the validation error and try again.
                if verbose:
                    logger.info(e)
                messages.extend([response, self.get_validation_error_message(e)])



"""LlamaBot ChatUI that composes with Bots."""

from typing import Callable, Optional


class ChatUIMixin:
    """A mixin for a chat user interface."""

    def __init__(
        self,
        initial_message: Optional[str] = None,
        callback_function: Optional[Callable] = None,
    ):
        import panel as pn

        self.callback_function = callback_function
        if callback_function is None:
            self.callback_function = lambda ai_message, user, instance: self(ai_message)

        self.chat_interface = pn.chat.ChatInterface(
            callback=self.callback_function, callback_exception="verbose"
        )
        if initial_message is not None:
            self.chat_interface.send(initial_message, user="System", respond=False)

    def servable(self):
        """Return the chat interface as a Panel servable object.

        :returns: The chat interface as a Panel servable object.
        """
        return self.chat_interface.servable()

    def serve(self, **kwargs):
        """Serve the chat interface.

        :returns: None
        """
        self.chat_interface.show(**kwargs)



"""Utility functions."""


def get_object_name(obj):
    """
    Get the name of the object as it's defined in the current namespace.

    :param obj: The object whose name we want to find.
    :return: The name of the object as a string, or None if not found.
    """
    for name, value in globals().items():
        if value is obj:
            return name
    return None



"""Bot abstractions that let me quickly build new GPT-based applications."""

from dotenv import load_dotenv

from llamabot.config import llamabotrc_path

load_dotenv()

if llamabotrc_path.exists():
    load_dotenv(llamabotrc_path)



"""Class definition for SimpleBot."""

import contextvars
from typing import Generator, Optional, Union


from llamabot.components.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
    BaseMessage,
)
from llamabot.recorder import autorecord, sqlite_log
from llamabot.config import default_language_model
from llamabot.cache import cache

prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class SimpleBot:
    """Simple Bot that is primed with a system prompt, accepts a human message,
    and sends back a single response.

    This bot does not retain chat history.

    :param system_prompt: The system prompt to use.
    :param temperature: The model temperature to use.
        See https://platform.openai.com/docs/api-reference/completions/create#completions/create-temperature
        for more information.
    :param model_name: The name of the model to use.
    :param stream_target: The target to stream the response to.
        Should be one of ("stdout", "panel", "api").
    :param json_mode: Whether to print debug messages.
    :param api_key: The OpenAI API key to use.
    :param mock_response: A mock response to use, for testing purposes only.
    :param completion_kwargs: Additional keyword arguments to pass to the completion function.
    """

    def __init__(
        self,
        system_prompt: str,
        temperature=0.0,
        model_name=default_language_model(),
        stream_target: str = "stdout",
        json_mode: bool = False,
        api_key: Optional[str] = None,
        mock_response: Optional[str] = None,
        **completion_kwargs,
    ):
        # Validate `stream_target.
        if stream_target not in ("stdout", "panel", "api", "none"):
            raise ValueError(
                f"stream_target must be one of ('stdout', 'panel', 'api', 'none'), got {stream_target}."
            )

        self.system_prompt = system_prompt
        if isinstance(system_prompt, str):
            self.system_prompt = SystemMessage(content=system_prompt)

        self.model_name = model_name
        self.stream_target = stream_target
        self.temperature = temperature
        self.json_mode = json_mode
        self.api_key = api_key
        self.mock_response = mock_response
        self.completion_kwargs = completion_kwargs

        # Set special cases for for o1 models.
        if model_name in ["o1-preview", "o1-mini"]:
            self.system_prompt = HumanMessage(content=system_prompt)
            self.temperature = 1.0
            self.stream_target = "none"

    def __call__(
        self, human_message: Union[str, BaseMessage]
    ) -> Union[AIMessage, Generator]:
        """Call the SimpleBot.

        :param human_message: The human message to use.
        :return: The response to the human message, primed by the system prompt.
        """
        if isinstance(human_message, str):
            human_message = HumanMessage(content=human_message)

        messages = [self.system_prompt, human_message]
        match self.stream_target:
            case "stdout":
                return self.stream_stdout(messages)
            case "panel":
                return self.stream_panel(messages)
            case "api":
                return self.stream_api(messages)
            case "none":
                return self.stream_none(messages)
        return AIMessage(content="")

    @cache.memoize(ignore={0})
    def stream_stdout(self, messages: list[BaseMessage]) -> AIMessage:
        """Stream the response to stdout.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages)
        message = ""
        for chunk in response:
            delta = chunk.choices[0].delta["content"]
            if delta is not None:
                print(delta, end="")
                message += delta
        response_message = AIMessage(content=message)
        autorecord(messages[-1].content, response_message.content)
        sqlite_log(self, messages + [response_message])
        return response_message

    def stream_panel(self, messages: list[BaseMessage]) -> Generator:
        """Stream the response to a Panel app.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages)
        response_message = ""
        for chunk in response:
            delta = chunk.choices[0].delta["content"]
            if delta is not None:
                response_message += delta
                yield response_message
        autorecord(messages[-1].content, response_message)
        sqlite_log(self, messages + [AIMessage(content=response_message)])

    @cache.memoize(ignore={0})
    def stream_none(self, messages: list[BaseMessage]) -> AIMessage:
        """Stream the response to None.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages, stream=False)
        response_message = AIMessage(
            content=response.choices[0].message.content.strip()
        )
        autorecord(messages[-1].content, response_message.content)
        sqlite_log(self, messages + [response_message])
        return response_message

    def stream_api(self, messages: list[BaseMessage]) -> Generator:
        """Stream the response to an API.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages)
        response_message = ""
        for chunk in response:
            delta = chunk.choices[0].delta["content"]
            if delta is not None:
                response_message += delta
                yield delta
        autorecord(messages[-1].content, response_message)
        sqlite_log(self, messages + [AIMessage(content=response_message)])

    @cache.memoize(ignore={0})
    def generate_response(self, messages: list[BaseMessage]) -> AIMessage:
        """Generate a response from the given messages.

        :param messages: A list of messages.
        :return: The response to the messages.
        """
        response = _make_response(self, messages)
        response_message = AIMessage(
            content=response.choices[0].message.content.strip()
        )
        autorecord(messages[-1].content, response_message.content)
        sqlite_log(self, messages + [response_message])
        return response_message

    def stream_response(self, messages: list[BaseMessage]):
        """Stream the response from the given messages.

        This is intended to be used with Panel's ChatInterface as part of the callback.

        :param messages: A list of messages.
        :return: A generator that yields the response.
        """
        response = _make_response(self, messages)
        message = ""
        for chunk in response:
            delta = chunk.choices[0].delta.content
            if delta is not None:
                message += delta
                print(delta, end="")
                yield message
        print()


def _make_response(bot: SimpleBot, messages: list[BaseMessage], stream: bool = True):
    """Make a response from the given messages.

    :param bot: A SimpleBot
    :param messages: A list of Messages.
    :return: A response object.
    """
    from litellm import completion

    messages_dumped: list[dict] = [m.model_dump() for m in messages]
    completion_kwargs = dict(
        model=bot.model_name,
        messages=messages_dumped,
        temperature=bot.temperature,
        stream=stream,
    )
    completion_kwargs.update(bot.completion_kwargs)

    if bot.mock_response:
        completion_kwargs["mock_response"] = bot.mock_response
    if bot.json_mode:
        completion_kwargs["response_format"] = {"type": "json_object"}
    if bot.api_key:
        completion_kwargs["api_key"] = bot.api_key
    return completion(**completion_kwargs)

    # Commented out until later.
    # def panel(
    #     self,
    #     input_text_label="Input",
    #     output_text_label="Output",
    #     submit_button_label="Submit",
    #     site_name="SimpleBot",
    #     title="SimpleBot",
    #     show=False,
    # ):
    #     """Create a Panel app that wraps a LlamaBot.

    #     :param input_text_label: The label for the input text.
    #     :param output_text_label: The label for the output text.
    #     :param submit_button_label: The label for the submit button.
    #     :param site_name: The name of the site.
    #     :param title: The title of the site.
    #     :param show: Whether to show the app.
    #         If False, we return the Panel app directly.
    #         If True, we call `.show()` on the app.
    #     :return: The Panel app, either showed or directly.
    #     """
    #     input_text = pn.widgets.TextAreaInput(
    #         name=input_text_label, value="", height=200, width=500
    #     )
    #     output_text = pn.pane.Markdown("")
    #     submit = pn.widgets.Button(name=submit_button_label, button_type="success")

    #     def b(event):
    #         """Button click handler.

    #         :param event: The button click event.
    #         """
    #         logger.info(input_text.value)
    #         output_text.object = ""
    #         markdown_handler = PanelMarkdownCallbackHandler(output_text)
    #         self.model.callback_manager.set_handler(markdown_handler)
    #         response = self(input_text.value)
    #         logger.info(response)

    #     submit.on_click(b)

    #     app = pn.template.FastListTemplate(
    #         site=site_name,
    #         title=title,
    #         main=[
    #             pn.Column(
    #                 *[
    #                     input_text,
    #                     submit,
    #                     pn.pane.Markdown(output_text_label),
    #                     output_text,
    #                 ]
    #             )
    #         ],
    #         main_max_width="768px",
    #     )
    #     app = pn.panel(app)
    #     if show:
    #         return app.show()
    #     return app



"""Top-level API for llamabot.

This is the file from which you can do:

    from llamabot import some_function

Use it to control the top-level API of your Python data science project.
"""

# Ensure that ~/.llamabotrc exists.
from pathlib import Path

from .bot.chatbot import ChatBot
from .bot.querybot import QueryBot
from .bot.simplebot import SimpleBot
from .bot.imagebot import ImageBot
from .bot.structuredbot import StructuredBot
from .recorder import PromptRecorder
from .prompt_manager import prompt

__all__ = [
    "ChatBot",
    "ImageBot",
    "SimpleBot",
    "QueryBot",
    "PromptRecorder",
    "StructuredBot",
    "prompt",
]


(Path.home() / ".llamabot").mkdir(parents=True, exist_ok=True)



"""Create a FastAPI app to visualize and compare prompts and messages."""

from typing import Optional
from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from pathlib import Path
import json
from pyprojroot import here
import logging

from llamabot.recorder import MessageLog, Base, upgrade_database, Prompt

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

templates = Jinja2Templates(directory="llamabot/web/templates")


def create_app(db_path: Optional[Path] = None):
    """Create a FastAPI app to visualize and compare prompts and messages.

    :param db_path: The path to the database to use.
    """
    if db_path is None:
        db_path = here() / "message_log.db"

    app = FastAPI()

    engine = create_engine(f"sqlite:///{db_path}")
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

    # Ensure the database is initialized and upgraded
    Base.metadata.create_all(engine)
    upgrade_database(engine)

    app.mount("/static", StaticFiles(directory="llamabot/web/static"), name="static")

    @app.get("/", response_class=HTMLResponse)
    async def root(request: Request):
        """The root page.

        :param request: The FastAPI request object.
        :return: The root page.
        """
        return templates.TemplateResponse("index.html", {"request": request})

    @app.get("/logs")
    async def get_logs():
        """Get all logs."""
        db = SessionLocal()
        try:
            logs = db.query(MessageLog).order_by(MessageLog.timestamp.desc()).all()
            logger.debug(f"Found {len(logs)} logs")

            log_data = []
            for log in logs:
                message_log = json.loads(log.message_log)
                prompt_hashes = set()
                for message in message_log:
                    if "prompt_hash" in message:
                        prompt_hashes.add(message["prompt_hash"])

                logger.debug(f"Log {log.id} has prompt hashes: {prompt_hashes}")

                prompts = db.query(Prompt).filter(Prompt.hash.in_(prompt_hashes)).all()
                logger.debug(f"Found prompts: {[p.function_name for p in prompts]}")

                prompt_names = [
                    f"- {prompt.function_name} ({prompt.hash[:6]})"
                    for prompt in prompts
                ]
                formatted_prompt_names = (
                    "\n".join(prompt_names) if prompt_names else "No prompts used"
                )
                logger.debug(
                    f"Prompt names for log {log.id}:\n{formatted_prompt_names}"
                )

                log_data.append(
                    {
                        "id": log.id,
                        "object_name": log.object_name,
                        "timestamp": log.timestamp,
                        "model_name": log.model_name,
                        "temperature": log.temperature,
                        "prompt_names": formatted_prompt_names,
                        "full_content": log.message_log,
                    }
                )

            return templates.TemplateResponse(
                "log_table.html",
                {
                    "request": {},
                    "logs": log_data,
                },
            )
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))
        finally:
            db.close()

    @app.get("/log/{log_id}")
    async def get_log(log_id: int):
        """Get a single log by ID."""
        db = SessionLocal()
        try:
            log = db.query(MessageLog).filter(MessageLog.id == log_id).first()
            if log is None:
                raise HTTPException(status_code=404, detail="Log not found")
            message_log = json.loads(log.message_log)

            # Fetch prompt names and templates for each message with a prompt_hash
            for message in message_log:
                if message.get("prompt_hash"):
                    prompt = (
                        db.query(Prompt)
                        .filter(Prompt.hash == message["prompt_hash"])
                        .first()
                    )
                    if prompt:
                        message["prompt_name"] = prompt.function_name
                        message["prompt_template"] = prompt.template

            return {
                "id": log.id,
                "object_name": log.object_name,
                "timestamp": log.timestamp,
                "message_log": message_log,
                "model_name": log.model_name,
                "temperature": log.temperature,
            }
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))
        finally:
            db.close()

    return app



"""Bots and prompts for semantic line breaks."""

from llamabot import SimpleBot
from llamabot.prompt_manager import prompt


@prompt(role="system")
def sembr_bot_system_prompt():
    """You are a SEMBR (semantic line breaks) bot. These are the SEMBR specification:

    Text written as plain text or a compatible markup language
    may use semantic line breaks.
    A semantic line break must not alter the final rendered output of the document.
    A semantic line break should not alter the intended meaning of the text.
    A semantic line break must occur after a sentence, as punctuated by a period (.),
    exclamation mark (!), or question mark (?).
    A semantic line break should occur after an independent clause
    as punctuated by a comma (,), semicolon (;), colon (:), or em dash (—).
    A semantic line break may occur after a dependent clause
    in order to clarify grammatical structure or satisfy line length constraints.
    A semantic line break is recommended before an enumerated or itemized list.
    A semantic line break may be used after one or more items
    in a list in order to logically group related items
    or satisfy line length constraints.
    A semantic line break must not occur within a hyphenated word.
    A semantic line break may occur before and after a hyperlink.
    A semantic line break may occur before inline markup.
    A maximum line length of 80 characters is recommended.
    A line may exceed the maximum line length if necessary,
    such as to accommodate hyperlinks, code elements, or other markup.
    """


def sembr_bot():
    """Return the SEMBR bot.

    :return: The SEMBR bot.
    """
    bot = SimpleBot(sembr_bot_system_prompt())
    return bot


@prompt(role="user")
def sembr(text):
    """Here is some text:

    {{ text }}

    Please insert "\n" characters according to the SEMBR specification.

    [BEGIN REWRITTEN TEXT]

    #noqa: DAR101
    """



"""Tests for llamabot."""

from llamabot import SimpleBot


def test_ollama_bot():
    """Test that SimpleBot works with Ollama."""
    bot = SimpleBot("You are a simple bot", model_name="ollama/gemma2:2b")
    bot("Hello there!")



"""Cache-related commands."""

import typer
from diskcache import Cache

app = typer.Typer()


@app.command()
def clear():
    """
    Clear the disk cache.

    This command will remove all items stored in the disk cache.
    """
    cache = Cache()
    cache.clear()
    typer.echo("Cache cleared successfully.")



"""Prompt recorder class definition."""

import contextvars
import json
import hashlib
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

from pyprojroot import here
from sqlalchemy import (
    Column,
    Connection,
    Engine,
    Float,
    Integer,
    String,
    Text,
    ForeignKey,
    create_engine,
    inspect,
    text,
)
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from tenacity import retry, stop_after_attempt, wait_exponential

from llamabot.components.messages import BaseMessage
from llamabot.utils import get_object_name

prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class PromptRecorder:
    """Prompt recorder to support recording of prompts and responses."""

    def __init__(self):
        self.prompts_and_responses = []

    def __enter__(self):
        """Enter the context manager.

        :returns: The prompt recorder.
        """
        prompt_recorder_var.set(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager.

        :param exc_type: The exception type.
        :param exc_val: The exception value.
        :param exc_tb: The exception traceback.
        """
        prompt_recorder_var.set(None)

    def log(self, prompt: str, response: str):
        """Log the prompt and response in chat history.

        :param prompt: The human prompt.
        :param response: A the response from the bot.
        """
        self.prompts_and_responses.append({"prompt": prompt, "response": response})

    def __repr__(self):
        """Return a string representation of the prompt recorder.

        :return: A string form of the prompts and responses as a dataframe.
        """
        import pandas as pd

        return pd.DataFrame(self.prompts_and_responses).__str__()

    def _repr_html_(self):
        """Return an HTML representation of the prompt recorder.

        :return: We delegate to the _repr_html_ method of the pandas DataFrame class.
        """
        return self.dataframe()._repr_html_()

    def dataframe(self):
        """Return a pandas DataFrame representation of the prompt recorder.

        :return: A pandas DataFrame representation of the prompt recorder.
        """
        import pandas as pd

        return pd.DataFrame(self.prompts_and_responses)

    def save(self, path: Path):
        """Save the prompt recorder to a path.

        :param path: The path to save the prompt recorder to.
        """
        path = Path(path)  # coerce to pathlib.Path
        with path.open("w+") as f:
            for prompt_and_response in self.prompts_and_responses:
                f.write(
                    f"**{prompt_and_response['prompt']}**\n\n{prompt_and_response['response']}\n\n"
                )

    def panel(self):
        """Return a panel representation of the prompt recorder.

        :return: A panel representation of the prompt recorder.
        """
        import panel as pn

        global index
        index = 0
        pn.extension()

        next_button = pn.widgets.Button(name=">")
        prev_button = pn.widgets.Button(name="<")

        buttons = pn.Row(prev_button, next_button)

        prompt_header = pn.pane.Markdown("# Prompt")
        prompt_display = pn.pane.Markdown(self.prompts_and_responses[index]["prompt"])

        prompt = pn.Column(prompt_header, prompt_display)

        response_header = pn.pane.Markdown("# Response")
        response_display = pn.pane.Markdown(
            self.prompts_and_responses[index]["response"]
        )
        response = pn.Column(response_header, response_display)

        display = pn.Row(prompt, response)

        def update_objects(index):
            """Update the prompt and response Markdown panes.

            :param index: The index of the prompt and response to update.
            """
            prompt_display.object = self.prompts_and_responses[index]["prompt"]
            response_display.object = self.prompts_and_responses[index]["response"]

        def next_button_callback(event):
            """Callback function for the next button.

            :param event: The click event.
            """
            global index
            index += 1
            if index > len(self.prompts_and_responses) - 1:
                index = len(self.prompts_and_responses) - 1

            update_objects(index)

        def prev_button_callback(event):
            """Callback function for the previous button.

            :param event: The click event.
            """
            global index
            index -= 1
            if index < 0:
                index = 0
            update_objects(index)

        next_button.on_click(next_button_callback)
        prev_button.on_click(prev_button_callback)

        return pn.Column(buttons, display)


def autorecord(prompt: str, response: str):
    """Record a prompt and response.

    This is intended to be called within every bot.
    If we are within a prompt recorder context,
    then the prompt recorder will record the prompt and response
    as specified in the function.

    :param prompt: The human prompt.
    :param response: A the response from the bot.
    """
    # Log the response.
    prompt_recorder: Optional[PromptRecorder] = prompt_recorder_var.get(None)
    if prompt_recorder:
        prompt_recorder.log(prompt, response)


# Design sqlite database for storing prompts and responses.
# Columns:
# - Python object name from globals()
# - Time stamp
# - Full message log as a JSON string of dictionaries.


Base = declarative_base()


class MessageLog(Base):
    """A log of a message exchange."""

    __tablename__ = "message_log"

    id = Column(Integer, primary_key=True)
    object_name = Column(String)
    timestamp = Column(String)
    message_log = Column(Text)
    model_name = Column(String)
    temperature = Column(Float, nullable=True)


def upgrade_database(engine: Engine):
    """Upgrade the database schema."""
    Base.metadata.create_all(engine)

    with engine.connect() as connection:
        inspector = inspect(engine)
        for table in [MessageLog, Prompt]:
            if inspector.has_table(table.__tablename__):
                existing_columns = [
                    c["name"] for c in inspector.get_columns(table.__tablename__)
                ]
                for column in table.__table__.columns:
                    if column.name not in existing_columns:
                        try:
                            add_column(connection, table.__tablename__, column)
                        except OperationalError as e:
                            print(f"Error adding column {column.name}: {e}")


def add_column(connection: Connection, table_name: str, column: Column):
    """
    Add a new column to an existing table.

    :param connection: SQLAlchemy connection
    :param table_name: Name of the table to modify
    :param column: Column object to add
    """
    column_name = column.name
    column_type = column.type.compile(connection.dialect)
    connection.execute(
        text(f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}")
    )


# Example usage:
# To add a new column in the future, you would do:
# add_column(engine, PromptResponseLog.__tablename__, Column('new_column_name', String))


@retry(stop=stop_after_attempt(5), wait=wait_exponential(multiplier=1, min=4, max=10))
def sqlite_log(obj: Any, messages: list[BaseMessage], db_path: Optional[Path] = None):
    """Log messages to the sqlite database for further analysis.

    :param obj: The object to log the messages for.
    :param messages: The messages to log.
    :param db_path: The path to the database to use.
        If not specified, defaults to ~/.llamabot/message_log.db
    """
    # Set up the database path
    if db_path is None:
        try:
            repo_root = here()
            db_path = repo_root / "message_log.db"

            # Ensure message_log.db is in .gitignore
            gitignore_path = repo_root / ".gitignore"
            if gitignore_path.exists():
                with open(gitignore_path, "r+") as f:
                    content = f.read()
                    if "message_log.db" not in content:
                        f.write("\n# SQLite database\nmessage_log.db\n")
            else:
                with open(gitignore_path, "w") as f:
                    f.write("# SQLite database\nmessage_log.db\n")
        except Exception:
            # If we're not in a git repo, use the home directory
            db_path = Path.home() / ".llamabot" / "message_log.db"

    # Create the engine and session
    engine = create_engine(f"sqlite:///{db_path}")
    Base.metadata.create_all(engine)
    upgrade_database(engine)  # Add this line to ensure the database is upgraded
    Session = sessionmaker(bind=engine)
    session = Session()

    # Get the object name
    object_name = get_object_name(obj)

    # Get the current timestamp
    timestamp = datetime.now().isoformat()

    # Convert messages to a JSON string, including prompt_hash
    message_log = json.dumps(
        [
            {
                "role": message.role,
                "content": message.content,
                "prompt_hash": (
                    message.prompt_hash if hasattr(message, "prompt_hash") else None
                ),
            }
            for message in messages
        ]
    )

    # Create a new MessageLog instance
    new_log = MessageLog(
        object_name=object_name,
        timestamp=timestamp,
        message_log=message_log,
        model_name=obj.model_name,
        temperature=obj.temperature if hasattr(obj, "temperature") else None,
    )

    # Add the new log to the session and commit
    session.add(new_log)
    session.commit()

    # Close the session
    session.close()


class Prompt(Base):
    """A version-controlled prompt template."""

    __tablename__ = "prompts"

    id = Column(Integer, primary_key=True)
    hash = Column(String, unique=True, index=True)
    template = Column(Text)
    function_name = Column(String)  # Add this line
    previous_version_id = Column(Integer, ForeignKey("prompts.id"), nullable=True)
    previous_version = relationship("Prompt", remote_side=[id])


def hash_template(template: str) -> str:
    """Hash a prompt template."""
    return hashlib.sha256(template.encode()).hexdigest()


def store_prompt_version(
    session, template: str, function_name: str, previous_hash: Optional[str] = None
):
    """Store a new prompt version in the database."""
    template_hash = hash_template(template)

    existing_prompt = session.query(Prompt).filter_by(hash=template_hash).first()
    if existing_prompt:
        return existing_prompt

    if previous_hash:
        previous_version = session.query(Prompt).filter_by(hash=previous_hash).first()
        previous_version_id = previous_version.id if previous_version else None
    else:
        previous_version_id = None

    new_prompt = Prompt(
        hash=template_hash,
        template=template,
        function_name=function_name,  # Add this line
        previous_version_id=previous_version_id,
    )
    session.add(new_prompt)
    session.commit()
    return new_prompt
